@@ -21,6 +21,7 @@
 */
 
 #include <sys/param.h>
+#include <sys/socket.h>
 #include <netdb.h>
 #include <sys/time.h>
 #include <time.h>
@@ -434,6 +435,15 @@ string uitoa(unsigned int i) // MSVC 6 doesn't grok overloading (un)signed
   return o.str();
 }
 
+string bitFlip(const string &str)
+{
+  string::size_type pos=0, epos=str.size();
+  string ret;
+  ret.reserve(epos);
+  for(;pos < epos; ++pos)
+    ret.append(1, ~str[pos]);
+  return ret;
+}
 
 string stringerror()
 {
@@ -671,7 +681,9 @@ int makeIPv6sockaddr(const std::string& addr, struct sockaddr_in6* ret)
     ourAddr.assign(addr.c_str() + 1, pos-1);
     port = atoi(addr.c_str()+pos+2);  
   }
-  
+  ret->sin6_scope_id=0;
+  ret->sin6_family=AF_INET6;
+
   if(inet_pton(AF_INET6, ourAddr.c_str(), (void*)&ret->sin6_addr) != 1) {
     struct addrinfo* res;
     struct addrinfo hints;
@@ -772,7 +784,7 @@ Regex::Regex(const string &expr)
     throw PDNSException("Regular expression did not compile");
 }
 
-void addCMsgSrcAddr(struct msghdr* msgh, void* cmsgbuf, ComboAddress* source)
+void addCMsgSrcAddr(struct msghdr* msgh, void* cmsgbuf, const ComboAddress* source)
 {
   struct cmsghdr *cmsg = NULL;
 
@@ -898,3 +910,27 @@ uint32_t burtle(const unsigned char* k, uint32_t length, uint32_t initval)
   /*-------------------------------------------- report the result */
   return c;
 }
+
+void setSocketTimestamps(int fd)
+{
+#ifdef SO_TIMESTAMP
+  int on=1;
+  if (setsockopt(fd, SOL_SOCKET, SO_TIMESTAMP, (char*)&on, sizeof(on)) < 0 )
+    L<<Logger::Error<<"Warning: unable to enable timestamp reporting for socket"<<endl;
+#endif
+}
+
+uint32_t pdns_strtoui(const char *nptr, char **endptr, int base)
+{
+#if ULONG_MAX == 4294967295
+  return strtoul(nptr, endptr, base);
+#else
+  unsigned long val = strtoul(nptr, endptr, base);
+  if (val > UINT_MAX) {
+   errno = ERANGE;
+   return UINT_MAX;
+  } 
+
+  return val;
+#endif
+}
