@@ -1,6 +1,6 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2003 - 2014  PowerDNS.COM BV
+    Copyright (C) 2003 - 2015  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2 
@@ -81,7 +81,10 @@ bool g_logCommonErrors;
 bool g_anyToTcp;
 uint16_t g_udpTruncationThreshold;
 __thread shared_ptr<RecursorLua>* t_pdl;
-__thread RemoteKeeper* t_remotes;
+
+__thread addrringbuf_t* t_remotes, *t_servfailremotes, *t_largeanswerremotes;
+
+__thread boost::circular_buffer<pair<std::string, uint16_t> >* t_queryring, *t_servfailqueryring;
 __thread shared_ptr<Regex>* t_traceRegex;
 
 RecursorControlChannel s_rcc; // only active in thread 0
@@ -118,15 +121,22 @@ typedef vector<int> tcpListenSockets_t;
 tcpListenSockets_t g_tcpListenSockets;   // shared across threads, but this is fine, never written to from a thread. All threads listen on all sockets
 int g_tcpTimeout;
 unsigned int g_maxMThreads;
-struct timeval g_now; // timestamp, updated (too) frequently
+__thread struct timeval g_now; // timestamp, updated (too) frequently
 typedef map<int, ComboAddress> listenSocketsAddresses_t; // is shared across all threads right now
 listenSocketsAddresses_t g_listenSocketsAddresses; // is shared across all threads right now
 
 __thread MT_t* MT; // the big MTasker
 
-unsigned int g_numThreads;
+unsigned int g_numThreads, g_numWorkerThreads;
 
-#define LOCAL_NETS "127.0.0.0/8, 10.0.0.0/8, 100.64.0.0/10, 169.254.0.0/16, 192.168.0.0/16, 172.16.0.0/12, ::1/128, fe80::/10"
+#define LOCAL_NETS "127.0.0.0/8, 10.0.0.0/8, 100.64.0.0/10, 169.254.0.0/16, 192.168.0.0/16, 172.16.0.0/12, ::1/128, fc00::/7, fe80::/10"
+// Bad Nets taken from both:
+// http://www.iana.org/assignments/iana-ipv4-special-registry/iana-ipv4-special-registry.xhtml 
+// and
+// http://www.iana.org/assignments/iana-ipv6-special-registry/iana-ipv6-special-registry.xhtml
+// where such a network may not be considered a valid destination
+#define BAD_NETS   "0.0.0.0/8, 192.0.0.0/24, 192.0.2.0/24, 198.51.100.0/24, 203.0.113.0/24, 240.0.0.0/4, ::/96, ::ffff:0:0/96, 100::/64, 2001:db8::/32"
+#define DONT_QUERY LOCAL_NETS ", " BAD_NETS
 
 //! used to send information to a newborn mthread
 struct DNSComboWriter {
@@ -139,6 +149,12 @@ struct DNSComboWriter {
     d_remote=*sa;
   }
 
+  void setLocal(const ComboAddress& sa)
+  {
+    d_local=sa;
+  }
+
+
   void setSocket(int sock)
   {
     d_socket=sock;
@@ -150,7 +166,7 @@ struct DNSComboWriter {
   }
 
   struct timeval d_now;
-  ComboAddress d_remote;
+  ComboAddress d_remote, d_local;
   bool d_tcp;
   int d_socket;
   shared_ptr<TCPConnection> d_tcpConnection;
@@ -269,9 +285,8 @@ static void setSocketSendBuffer(int fd, uint32_t size)
 class UDPClientSocks
 {
   unsigned int d_numsocks;
-  unsigned int d_maxsocks;
 public:
-  UDPClientSocks() : d_numsocks(0), d_maxsocks(5000)
+  UDPClientSocks() : d_numsocks(0)
   {
   }
 
@@ -329,7 +344,7 @@ public:
   // returns -1 for errors which might go away, throws for ones that won't
   static int makeClientSocket(int family)
   {
-    int ret=(int)socket(family, SOCK_DGRAM, 0);
+    int ret=(int)socket(family, SOCK_DGRAM, 0 ); // turns out that setting CLO_EXEC and NONBLOCK from here is not a performance win on Linux (oddly enough)
 
     if(ret < 0 && errno==EMFILE) // this is not a catastrophic error
       return ret;
@@ -478,10 +493,18 @@ TCPConnection::~TCPConnection()
 AtomicCounter TCPConnection::s_currentConnections; 
 void handleRunningTCPQuestion(int fd, FDMultiplexer::funcparam_t& var);
 
-void updateRcodeStats(int res)
+// the idea is, only do things that depend on the *response* here. Incoming accounting is on incoming.
+void updateResponseStats(int res, const ComboAddress& remote, unsigned int packetsize, const std::string* query, uint16_t qtype)
 {
+  if(packetsize > 1000 && t_largeanswerremotes)
+    t_largeanswerremotes->push_back(remote);
   switch(res) {
   case RCode::ServFail:
+    if(t_servfailremotes) {
+      t_servfailremotes->push_back(remote);
+      if(query) // packet cache
+	t_servfailqueryring->push_back(make_pair(*query, qtype));
+    }
     g_stats.servFails++;
     break;
   case RCode::NXDomain:
@@ -495,14 +518,22 @@ void updateRcodeStats(int res)
 
 ResponseStats g_rs;
 
+static string makeLoginfo(DNSComboWriter* dc)
+try
+{
+  return "("+dc->d_mdp.d_qname+"/"+DNSRecordContent::NumberToType(dc->d_mdp.d_qtype)+" from "+(dc->d_remote.toString())+")";
+}
+catch(...)
+{
+  return "Exception making error message for exception";
+}
+
 void startDoResolve(void *p)
 {
   DNSComboWriter* dc=(DNSComboWriter *)p;
-  string loginfo="";
-
   try {
-    loginfo=" (while setting loginfo)";
-    loginfo=" ("+dc->d_mdp.d_qname+"/"+lexical_cast<string>(dc->d_mdp.d_qtype)+" from "+(dc->d_remote.toString())+")";
+    t_queryring->push_back(make_pair(dc->d_mdp.d_qname, dc->d_mdp.d_qtype));
+
     uint32_t maxanswersize= dc->d_tcp ? 65535 : min((uint16_t) 512, g_udpTruncationThreshold);
     EDNSOpts edo;
     if(getEDNSOpts(dc->d_mdp, &edo) && !dc->d_tcp) {
@@ -525,6 +556,10 @@ void startDoResolve(void *p)
     uint32_t minTTL=std::numeric_limits<uint32_t>::max();
 
     SyncRes sr(dc->d_now);
+    if(t_pdl) {
+      sr.setLuaEngine(*t_pdl);
+      sr.d_requestor=dc->d_remote;
+    }
     bool tracedQuery=false; // we could consider letting Lua know about this too
     bool variableAnswer = false;
 
@@ -543,7 +578,7 @@ void startDoResolve(void *p)
     }
     
     if(!g_quiet || tracedQuery)
-      L<<Logger::Warning<<t_id<<" ["<<MT->getTid()<<"] " << (dc->d_tcp ? "TCP " : "") << "question for '"<<dc->d_mdp.d_qname<<"|"
+      L<<Logger::Warning<<t_id<<" ["<<MT->getTid()<<"/"<<MT->numProcesses()<<"] " << (dc->d_tcp ? "TCP " : "") << "question for '"<<dc->d_mdp.d_qname<<"|"
        <<DNSRecordContent::NumberToType(dc->d_mdp.d_qtype)<<"' from "<<dc->getRemote()<<endl;
 
     sr.setId(MT->getTid());
@@ -567,8 +602,8 @@ void startDoResolve(void *p)
         res = sr.beginResolve(dc->d_mdp.d_qname, QType(dc->d_mdp.d_qtype), dc->d_mdp.d_qclass, ret);
       }
       catch(ImmediateServFailException &e) {
-        L<<Logger::Error<<"Sending SERVFAIL during resolve of '"<<dc->d_mdp.d_qname<<"' because: "<<e.reason<<endl;
-
+        if(g_logCommonErrors)
+          L<<Logger::Notice<<"Sending SERVFAIL to "<<dc->getRemote()<<" during resolve of '"<<dc->d_mdp.d_qname<<"' because: "<<e.reason<<endl;
         res = RCode::ServFail;
       }
 
@@ -614,7 +649,7 @@ void startDoResolve(void *p)
     }
     else {
       pw.getHeader()->rcode=res;
-      updateRcodeStats(res);
+
       
       if(ret.size()) {
         orderAndShuffle(ret);
@@ -645,8 +680,17 @@ void startDoResolve(void *p)
     }
   sendit:;
     g_rs.submitResponse(dc->d_mdp.d_qtype, packet.size(), !dc->d_tcp);
+    updateResponseStats(res, dc->d_remote, packet.size(), &dc->d_mdp.d_qname, dc->d_mdp.d_qtype);
     if(!dc->d_tcp) {
-      sendto(dc->d_socket, (const char*)&*packet.begin(), packet.size(), 0, (struct sockaddr *)(&dc->d_remote), dc->d_remote.getSocklen());
+      struct msghdr msgh;
+      struct iovec iov;
+      char cbuf[256];
+      fillMSGHdr(&msgh, &iov, cbuf, 0, (char*)&*packet.begin(), packet.size(), &dc->d_remote);
+      if(dc->d_local.sin4.sin_family)
+	addCMsgSrcAddr(&msgh, cbuf, &dc->d_local);
+      else
+        msgh.msg_control=NULL;
+      sendmsg(dc->d_socket, &msgh, 0);
       if(!SyncRes::s_nopacketcache && !variableAnswer ) {
         t_packetCache->insertResponsePacket(string((const char*)&*packet.begin(), packet.size()),
                                             g_now.tv_sec, 
@@ -693,9 +737,10 @@ void startDoResolve(void *p)
     }
     
     if(!g_quiet) {
-      L<<Logger::Error<<t_id<<" ["<<MT->getTid()<<"] answer to "<<(dc->d_mdp.d_header.rd?"":"non-rd ")<<"question '"<<dc->d_mdp.d_qname<<"|"<<DNSRecordContent::NumberToType(dc->d_mdp.d_qtype);
+      L<<Logger::Error<<t_id<<" ["<<MT->getTid()<<"/"<<MT->numProcesses()<<"] answer to "<<(dc->d_mdp.d_header.rd?"":"non-rd ")<<"question '"<<dc->d_mdp.d_qname<<"|"<<DNSRecordContent::NumberToType(dc->d_mdp.d_qtype);
       L<<"': "<<ntohs(pw.getHeader()->ancount)<<" answers, "<<ntohs(pw.getHeader()->arcount)<<" additional, took "<<sr.d_outqueries<<" packets, "<<
-      sr.d_throttledqueries<<" throttled, "<<sr.d_timeouts<<" timeouts, "<<sr.d_tcpoutqueries<<" tcp connections, rcode="<<res<<endl;
+	sr.d_totUsec/1000.0<<" ms, "<<
+	sr.d_throttledqueries<<" throttled, "<<sr.d_timeouts<<" timeouts, "<<sr.d_tcpoutqueries<<" tcp connections, rcode="<<res<<endl;
     }
 
     sr.d_outqueries ? t_RC->cacheMisses++ : t_RC->cacheHits++; 
@@ -715,23 +760,24 @@ void startDoResolve(void *p)
     newLat = min(newLat,(uint64_t)(g_networkTimeoutMsec*1000)); // outliers of several minutes exist..
     g_stats.avgLatencyUsec=(1-1.0/g_latencyStatSize)*g_stats.avgLatencyUsec + (float)newLat/g_latencyStatSize;
     // no worries, we do this for packet cache hits elsewhere
+    //    cout<<dc->d_mdp.d_qname<<"\t"<<MT->getUsec()<<"\t"<<sr.d_outqueries<<endl;
     delete dc;
     dc=0;
   }
   catch(PDNSException &ae) {
-    L<<Logger::Error<<"startDoResolve problem"<<loginfo<<": "<<ae.reason<<endl;
+    L<<Logger::Error<<"startDoResolve problem "<<makeLoginfo(dc)<<": "<<ae.reason<<endl;
     delete dc;
   }
   catch(MOADNSException& e) {
-    L<<Logger::Error<<"DNS parser error"<<loginfo<<": "<<dc->d_mdp.d_qname<<", "<<e.what()<<endl;
+    L<<Logger::Error<<"DNS parser error "<<makeLoginfo(dc) <<": "<<dc->d_mdp.d_qname<<", "<<e.what()<<endl;
     delete dc;
   }
   catch(std::exception& e) {
-    L<<Logger::Error<<"STL error"<<loginfo<<": "<<e.what()<<endl;
+    L<<Logger::Error<<"STL error "<< makeLoginfo(dc)<<": "<<e.what()<<endl;
     delete dc;
   }
   catch(...) {
-    L<<Logger::Error<<"Any other exception in a resolver context"<<loginfo<<endl;
+    L<<Logger::Error<<"Any other exception in a resolver context "<< makeLoginfo(dc) <<endl;
   }
   
   g_stats.maxMThreadStackUsage = max(MT->getMaxStackUsage(), g_stats.maxMThreadStackUsage);
@@ -856,7 +902,8 @@ void handleNewTCPQuestion(int fd, FDMultiplexer::funcparam_t& )
       return;
     }
 
-    t_remotes->addRemote(addr);
+    if(t_remotes)
+      t_remotes->push_back(addr);
     if(t_allowFrom && !t_allowFrom->match(&addr)) {
       if(!g_quiet) 
         L<<Logger::Error<<"["<<MT->getTid()<<"] dropping TCP query from "<<addr.toString()<<", address not matched by allow-from"<<endl;
@@ -883,8 +930,17 @@ void handleNewTCPQuestion(int fd, FDMultiplexer::funcparam_t& )
   }
 }
  
-string* doProcessUDPQuestion(const std::string& question, const ComboAddress& fromaddr, int fd)
+string* doProcessUDPQuestion(const std::string& question, const ComboAddress& fromaddr, const ComboAddress& destaddr, struct timeval tv, int fd)
 {
+  gettimeofday(&g_now, 0);
+  struct timeval diff = g_now - tv;
+  double delta=(diff.tv_sec*1000 + diff.tv_usec/1000.0);
+
+  if(tv.tv_sec && delta > 1000.0) {
+    g_stats.tooOldDrops++;
+    return 0;
+  }
+
   ++g_stats.qcounter;
   if(fromaddr.sin4.sin_family==AF_INET6)
      g_stats.ipv6qcounter++;
@@ -894,16 +950,28 @@ string* doProcessUDPQuestion(const std::string& question, const ComboAddress& fr
     uint32_t age;
     if(!SyncRes::s_nopacketcache && t_packetCache->getResponsePacket(question, g_now.tv_sec, &response, &age)) {
       if(!g_quiet)
-        L<<Logger::Error<<t_id<< " question answered from packet cache from "<<fromaddr.toString()<<endl;
+        L<<Logger::Notice<<t_id<< " question answered from packet cache from "<<fromaddr.toString()<<endl;
+      // t_queryring->push_back("packetcached");
       
       g_stats.packetCacheHits++;
       SyncRes::s_queries++;
       ageDNSPacket(response, age);
-      sendto(fd, response.c_str(), response.length(), 0, (struct sockaddr*) &fromaddr, fromaddr.getSocklen());
+      struct msghdr msgh;
+      struct iovec iov;
+      char cbuf[256];
+      fillMSGHdr(&msgh, &iov, cbuf, 0, (char*)response.c_str(), response.length(), const_cast<ComboAddress*>(&fromaddr));
+      if(destaddr.sin4.sin_family) {
+	addCMsgSrcAddr(&msgh, cbuf, &destaddr);
+      }
+      else {
+        msgh.msg_control=NULL;
+      }
+      sendmsg(fd, &msgh, 0);
+
       if(response.length() >= sizeof(struct dnsheader)) {
         struct dnsheader dh;
         memcpy(&dh, response.c_str(), sizeof(dh));
-        updateRcodeStats(dh.rcode);
+        updateResponseStats(dh.rcode, fromaddr, response.length(), 0, 0);
       }
       g_stats.avgLatencyUsec=(1-1.0/g_latencyStatSize)*g_stats.avgLatencyUsec + 0.0; // we assume 0 usec
       return 0;
@@ -914,7 +982,19 @@ string* doProcessUDPQuestion(const std::string& question, const ComboAddress& fr
     return 0;
   }
   
+  if(t_pdl->get()) {
+    if((*t_pdl)->ipfilter(fromaddr, destaddr)) {
+      if(!g_quiet)
+	L<<Logger::Notice<<t_id<<" ["<<MT->getTid()<<"/"<<MT->numProcesses()<<"] DROPPED question from "<<fromaddr.toStringWithPort()<<" based on policy"<<endl;
+      g_stats.policyDrops++;
+      return 0;
+    }
+  }
+
   if(MT->numProcesses() > g_maxMThreads) {
+    if(!g_quiet)
+      L<<Logger::Notice<<t_id<<" ["<<MT->getTid()<<"/"<<MT->numProcesses()<<"] DROPPED question from "<<fromaddr.toStringWithPort()<<", over capacity"<<endl;
+
     g_stats.overCapacityDrops++;
     return 0;
   }
@@ -922,21 +1002,30 @@ string* doProcessUDPQuestion(const std::string& question, const ComboAddress& fr
   DNSComboWriter* dc = new DNSComboWriter(question.c_str(), question.size(), g_now);
   dc->setSocket(fd);
   dc->setRemote(&fromaddr);
+  dc->setLocal(destaddr);
 
   dc->d_tcp=false;
   MT->makeThread(startDoResolve, (void*) dc); // deletes dc
   return 0;
 } 
  
+
 void handleNewUDPQuestion(int fd, FDMultiplexer::funcparam_t& var)
 {
   int len;
   char data[1500];
   ComboAddress fromaddr;
-  socklen_t addrlen=sizeof(fromaddr);
-  
-  if((len=recvfrom(fd, data, sizeof(data), 0, (sockaddr *)&fromaddr, &addrlen)) >= 0) {
-    t_remotes->addRemote(fromaddr);
+  struct msghdr msgh;
+  struct iovec iov;
+  char cbuf[256];
+
+  fromaddr.sin6.sin6_family=AF_INET6; // this makes sure fromaddr is big enough
+  fillMSGHdr(&msgh, &iov, cbuf, sizeof(cbuf), data, sizeof(data), &fromaddr);
+
+  for(;;) 
+  if((len=recvmsg(fd, &msgh, 0)) >= 0) {
+    if(t_remotes)
+      t_remotes->push_back(fromaddr);
 
     if(t_allowFrom && !t_allowFrom->match(&fromaddr)) {
       if(!g_quiet) 
@@ -966,10 +1055,15 @@ void handleNewUDPQuestion(int fd, FDMultiplexer::funcparam_t& var)
       }
       else {
         string question(data, len);
+	struct timeval tv={0,0};
+	HarvestTimestamp(&msgh, &tv);
+	ComboAddress dest;
+	memset(&dest, 0, sizeof(dest)); // this makes sure we igore this address if not returned by recvmsg above
+	HarvestDestinationAddress(&msgh, &dest);
         if(g_weDistributeQueries)
-          distributeAsyncFunction(question, boost::bind(doProcessUDPQuestion, question, fromaddr, fd));
+          distributeAsyncFunction(question, boost::bind(doProcessUDPQuestion, question, fromaddr, dest, tv, fd));
         else
-          doProcessUDPQuestion(question, fromaddr, fd);
+          doProcessUDPQuestion(question, fromaddr, dest, tv, fd);
       }
     }
     catch(MOADNSException& mde) {
@@ -980,8 +1074,9 @@ void handleNewUDPQuestion(int fd, FDMultiplexer::funcparam_t& var)
   }
   else {
     // cerr<<t_id<<" had error: "<<stringerror()<<endl;
-    if(errno == EAGAIN)
+    if(errno == EAGAIN) 
       g_stats.noPacketError++;
+    break;
   }
 }
 
@@ -1014,17 +1109,20 @@ void makeTCPServerSockets()
     }
 
     fd=socket(sin.sin6.sin6_family, SOCK_STREAM, 0);
-    Utility::setCloseOnExec(fd);
-
     if(fd<0) 
       throw PDNSException("Making a TCP server socket for resolver: "+stringerror());
 
+    Utility::setCloseOnExec(fd);
+
     int tmp=1;
     if(setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,(char*)&tmp,sizeof tmp)<0) {
       L<<Logger::Error<<"Setsockopt failed for TCP listening socket"<<endl;
       exit(1);
     }
-    
+    if(sin.sin6.sin6_family == AF_INET6 && setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &tmp, sizeof(tmp)) < 0) {
+      L<<Logger::Error<<"Failed to set IPv6 socket to IPv6 only, continuing anyhow: "<<strerror(errno)<<endl;
+    }
+
 #ifdef TCP_DEFER_ACCEPT
     if(setsockopt(fd, SOL_TCP,TCP_DEFER_ACCEPT,(char*)&tmp,sizeof tmp) >= 0) {
       if(i==locals.begin())
@@ -1059,10 +1157,6 @@ void makeUDPServerSockets()
   if(locals.empty())
     throw PDNSException("No local address specified");
   
-  if(::arg()["local-address"]=="0.0.0.0") {
-    L<<Logger::Warning<<"It is advised to bind to explicit addresses with the --local-address option"<<endl;
-  }
-
   for(vector<string>::const_iterator i=locals.begin();i!=locals.end();++i) {
     ServiceTuple st;
     st.port=::arg().asNum("local-port");
@@ -1079,11 +1173,24 @@ void makeUDPServerSockets()
     }
     
     int fd=socket(sin.sin4.sin_family, SOCK_DGRAM, 0);
-    Utility::setCloseOnExec(fd);
-
     if(fd < 0) {
       throw PDNSException("Making a UDP server socket for resolver: "+netstringerror());
     }
+    setSocketTimestamps(fd);
+
+    if(IsAnyAddress(sin)) {
+      int one=1;
+      setsockopt(fd, IPPROTO_IP, GEN_IP_PKTINFO, &one, sizeof(one));     // linux supports this, so why not - might fail on other systems
+#ifdef IPV6_RECVPKTINFO
+      setsockopt(fd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &one, sizeof(one)); 
+#endif
+      if(sin.sin6.sin6_family == AF_INET6 && setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &one, sizeof(one)) < 0) {
+	L<<Logger::Error<<"Failed to set IPv6 socket to IPv6 only, continuing anyhow: "<<strerror(errno)<<endl;
+      }
+
+    }
+
+    Utility::setCloseOnExec(fd);
 
     setSocketReceiveBuffer(fd, 250000);
     sin.sin4.sin_port = htons(st.port);
@@ -1122,7 +1229,7 @@ void daemonize(void)
   }
 }
 
-uint64_t counter;
+AtomicCounter counter;
 bool statsWanted;
 
 void usr1Handler(int)
@@ -1180,65 +1287,80 @@ void doStats(void)
 }
 
 static void houseKeeping(void *)
-try
 {
   static __thread time_t last_stat, last_rootupdate, last_prune, last_secpoll;
   static __thread int cleanCounter=0;
-  struct timeval now;
-  Utility::gettimeofday(&now, 0);
-
-  // clog<<"* "<<t_id<<" "<<(void*)&last_stat<<"\t"<<(unsigned int)last_stat<<endl;
-
-  if(now.tv_sec - last_prune > (time_t)(5 + t_id)) { 
-    DTime dt;
-    dt.setTimeval(now);
-    t_RC->doPrune(); // this function is local to a thread, so fine anyhow
-    t_packetCache->doPruneTo(::arg().asNum("max-packetcache-entries") / g_numThreads);
+  static __thread bool s_running;  // houseKeeping can get suspended in secpoll, and be restarted, which makes us do duplicate work
+  try {
+    if(s_running)
+      return;
+    s_running=true;
     
-    pruneCollection(t_sstorage->negcache, ::arg().asNum("max-cache-entries") / (g_numThreads * 10), 200);
+    struct timeval now;
+    Utility::gettimeofday(&now, 0);
     
-    if(!((cleanCounter++)%40)) {  // this is a full scan!
-      time_t limit=now.tv_sec-300;
-      for(SyncRes::nsspeeds_t::iterator i = t_sstorage->nsSpeeds.begin() ; i!= t_sstorage->nsSpeeds.end(); )
-        if(i->second.stale(limit))
-          t_sstorage->nsSpeeds.erase(i++);
-        else
-          ++i;
+    if(now.tv_sec - last_prune > (time_t)(5 + t_id)) { 
+      DTime dt;
+      dt.setTimeval(now);
+      t_RC->doPrune(); // this function is local to a thread, so fine anyhow
+      t_packetCache->doPruneTo(::arg().asNum("max-packetcache-entries") / g_numWorkerThreads);
+      
+      pruneCollection(t_sstorage->negcache, ::arg().asNum("max-cache-entries") / (g_numWorkerThreads * 10), 200);
+      
+      if(!((cleanCounter++)%40)) {  // this is a full scan!
+	time_t limit=now.tv_sec-300;
+	for(SyncRes::nsspeeds_t::iterator i = t_sstorage->nsSpeeds.begin() ; i!= t_sstorage->nsSpeeds.end(); )
+	  if(i->second.stale(limit))
+	    t_sstorage->nsSpeeds.erase(i++);
+	  else
+	    ++i;
+      }
+      last_prune=time(0);
     }
-//    L<<Logger::Warning<<"Spent "<<dt.udiff()/1000<<" msec cleaning"<<endl;
-    last_prune=time(0);
+    
+    if(now.tv_sec - last_rootupdate > 7200) {
+      SyncRes sr(now);
+      sr.setDoEDNS0(true);
+      vector<DNSResourceRecord> ret;
+      
+      sr.setNoCache();
+      int res=-1;
+      try {
+	res=sr.beginResolve(".", QType(QType::NS), 1, ret);
+      }
+      catch(...)
+	{
+	  L<<Logger::Error<<"Failed to update . records, got an exception"<<endl;
+	}
+      if(!res) {
+	L<<Logger::Notice<<"Refreshed . records"<<endl;
+	last_rootupdate=now.tv_sec;
+      }
+      else
+	L<<Logger::Error<<"Failed to update . records, RCODE="<<res<<endl;
+    }
+    
+    if(!t_id) {
+      if(now.tv_sec - last_stat >= 1800) { 
+	doStats();
+	last_stat=time(0);
+      }
+      
+      if(now.tv_sec - last_secpoll >= 3600) {
+	try {
+	  doSecPoll(&last_secpoll);
+	}
+	catch(...) {}
+      }
+    }
+    s_running=false;
   }
-  
-  if(now.tv_sec - last_rootupdate > 7200) {
-    SyncRes sr(now);
-    sr.setDoEDNS0(true);
-    vector<DNSResourceRecord> ret;
-
-    sr.setNoCache();
-    int res=sr.beginResolve(".", QType(QType::NS), 1, ret);
-    if(!res) {
-      L<<Logger::Warning<<"Refreshed . records"<<endl;
-      last_rootupdate=now.tv_sec;
+  catch(PDNSException& ae)
+    {
+      s_running=false;
+      L<<Logger::Error<<"Fatal error in housekeeping thread: "<<ae.reason<<endl;
+      throw;
     }
-    else
-      L<<Logger::Error<<"Failed to update . records, RCODE="<<res<<endl;
-  }
-
-  if(!t_id) {
-    if(now.tv_sec - last_stat >= 1800) { 
-      doStats();
-      last_stat=time(0);
-    }
-
-    if(now.tv_sec - last_secpoll >= 3600) {
-      doSecPoll(&last_secpoll);
-    }
-  }
-}
-catch(PDNSException& ae)
-{
-  L<<Logger::Error<<"Fatal error in housekeeping thread: "<<ae.reason<<endl;
-  throw;
 }
 
 void makeThreadPipes()
@@ -1350,6 +1472,13 @@ vector<ComboAddress>& operator+=(vector<ComboAddress>&a, const vector<ComboAddre
   return a;
 }
 
+vector<pair<string, uint16_t> >& operator+=(vector<pair<string, uint16_t> >&a, const vector<pair<string, uint16_t> >& b)
+{
+  a.insert(a.end(), b.begin(), b.end());
+  return a;
+}
+
+
 template<class T> T broadcastAccFunction(const boost::function<T*()>& func, bool skipSelf)
 {
   unsigned int n = 0;
@@ -1392,6 +1521,7 @@ template<class T> T broadcastAccFunction(const boost::function<T*()>& func, bool
 template string broadcastAccFunction(const boost::function<string*()>& fun, bool skipSelf); // explicit instantiation
 template uint64_t broadcastAccFunction(const boost::function<uint64_t*()>& fun, bool skipSelf); // explicit instantiation
 template vector<ComboAddress> broadcastAccFunction(const boost::function<vector<ComboAddress> *()>& fun, bool skipSelf); // explicit instantiation
+template vector<pair<string,uint16_t> > broadcastAccFunction(const boost::function<vector<pair<string, uint16_t> > *()>& fun, bool skipSelf); // explicit instantiation
 
 void handleRCC(int fd, FDMultiplexer::funcparam_t& var)
 {
@@ -1520,7 +1650,7 @@ void handleUDPServerResponse(int fd, FDMultiplexer::funcparam_t& var)
   pident.fd=fd;
 
   if(!dh.qr && g_logCommonErrors) {
-    L<<Logger::Warning<<"Not taking data from question on outgoing socket from "<< fromaddr.toStringWithPort()  <<endl;
+    L<<Logger::Notice<<"Not taking data from question on outgoing socket from "<< fromaddr.toStringWithPort()  <<endl;
   }
 
   if(!dh.qdcount || // UPC, Nominum, very old BIND on FormErr, NSD
@@ -1662,20 +1792,19 @@ static void checkLinuxIPv6Limits()
 }
 static void checkOrFixFDS()
 {
-  unsigned int availFDs=getFilenumLimit();
-  if(g_maxMThreads * g_numThreads > availFDs) {
-    if(getFilenumLimit(true) >= g_maxMThreads * g_numThreads) {
-      setFilenumLimit(g_maxMThreads * g_numThreads);
-      L<<Logger::Warning<<"Raised soft limit on number of filedescriptors to "<<g_maxMThreads * g_numThreads<<" to match max-mthreads and threads settings"<<endl;
+  unsigned int availFDs=getFilenumLimit()-10; // some healthy margin, thanks AJ ;-)
+  if(g_maxMThreads * g_numWorkerThreads > availFDs) {
+    if(getFilenumLimit(true) >= g_maxMThreads * g_numWorkerThreads) {
+      setFilenumLimit(g_maxMThreads * g_numWorkerThreads);
+      L<<Logger::Warning<<"Raised soft limit on number of filedescriptors to "<<g_maxMThreads * g_numWorkerThreads<<" to match max-mthreads and threads settings"<<endl;
     }
     else {
-      int newval = getFilenumLimit(true) / g_numThreads;
-      L<<Logger::Warning<<"Insufficient number of filedescriptors available for max-mthreads*threads setting! ("<<availFDs<<" < "<<g_maxMThreads*g_numThreads<<"), reducing max-mthreads to "<<newval<<endl;
+      int newval = getFilenumLimit(true) / g_numWorkerThreads;
+      L<<Logger::Warning<<"Insufficient number of filedescriptors available for max-mthreads*threads setting! ("<<availFDs<<" < "<<g_maxMThreads*g_numWorkerThreads<<"), reducing max-mthreads to "<<newval<<endl;
       g_maxMThreads = newval;
-      setFilenumLimit(g_maxMThreads * g_numThreads);
+      setFilenumLimit(g_maxMThreads * g_numWorkerThreads);
     }
   }
-
 }
 
 void* recursorThread(void*);
@@ -1698,7 +1827,7 @@ void parseACLs()
     cleanSlashes(configname);
     
     if(!::arg().preParseFile(configname.c_str(), "allow-from-file")) 
-      L<<Logger::Warning<<"Unable to re-parse configuration file '"<<configname<<"'"<<endl;
+      throw runtime_error("Unable to re-parse configuration file '"+configname+"'");
     ::arg().preParseFile(configname.c_str(), "allow-from", LOCAL_NETS);
     ::arg().preParseFile(configname.c_str(), "include-dir");
     ::arg().preParse(g_argc, g_argv, "include-dir");
@@ -1708,8 +1837,10 @@ void parseACLs()
     ::arg().gatherIncludes(extraConfigs);
 
     BOOST_FOREACH(const std::string& fn, extraConfigs) {
-      ::arg().preParseFile(fn.c_str(), "allow-from-file", ::arg()["allow-from-file"]);
-      ::arg().preParseFile(fn.c_str(), "allow-from", ::arg()["allow-from"]);
+      if(!::arg().preParseFile(fn.c_str(), "allow-from-file", ::arg()["allow-from-file"]))
+	throw runtime_error("Unable to re-parse configuration file include '"+fn+"'");
+      if(!::arg().preParseFile(fn.c_str(), "allow-from", ::arg()["allow-from"]))
+	throw runtime_error("Unable to re-parse configuration file include '"+fn+"'");
     }
 
     ::arg().preParse(g_argc, g_argv, "allow-from-file");
@@ -1785,7 +1916,8 @@ int serviceMain(int argc, char*argv[])
   g_disthashseed=dns_random(0xffffffff);
 
   parseACLs();
-  
+  sortPublicSuffixList();
+
   if(!::arg()["dont-query"].empty()) {
     g_dontQuery=new NetmaskGroup;
     vector<string> ips;
@@ -1804,6 +1936,7 @@ int serviceMain(int argc, char*argv[])
   }
 
   g_quiet=::arg().mustDo("quiet");
+  
   g_weDistributeQueries = ::arg().mustDo("pdns-distributes-queries");
   if(g_weDistributeQueries) {
       L<<Logger::Warning<<"PowerDNS Recursor itself will distribute queries over threads"<<endl;
@@ -1861,6 +1994,9 @@ int serviceMain(int argc, char*argv[])
   SyncRes::s_serverdownmaxfails=::arg().asNum("server-down-max-fails");
   SyncRes::s_serverdownthrottletime=::arg().asNum("server-down-throttle-time");
   SyncRes::s_serverID=::arg()["server-id"];
+  SyncRes::s_maxqperq=::arg().asNum("max-qperq");
+  SyncRes::s_maxtotusec=1000*::arg().asNum("max-total-msec");
+  SyncRes::s_rootNXTrust = ::arg().mustDo( "root-nx-trust");
   if(SyncRes::s_serverID.empty()) {
     char tmp[128];
     gethostname(tmp, sizeof(tmp)-1);
@@ -1901,6 +2037,9 @@ int serviceMain(int argc, char*argv[])
   signal(SIGPIPE,SIG_IGN);
   writePid();
   makeControlChannelSocket( ::arg().asNum("processes") > 1 ? forks : -1);
+  g_numThreads = ::arg().asNum("threads") + ::arg().mustDo("pdns-distributes-queries");
+  g_maxMThreads = ::arg().asNum("max-mthreads");
+  checkOrFixFDS();
   
   int newgid=0;
   if(!::arg()["setgid"].empty())
@@ -1920,13 +2059,11 @@ int serviceMain(int argc, char*argv[])
 
   Utility::dropUserPrivs(newuid);
   g_numThreads = ::arg().asNum("threads") + ::arg().mustDo("pdns-distributes-queries");
-  
+  g_numWorkerThreads = ::arg().asNum("threads");
   makeThreadPipes();
   
   g_tcpTimeout=::arg().asNum("client-tcp-timeout");
   g_maxTCPPerClient=::arg().asNum("max-tcp-per-client");
-  g_maxMThreads=::arg().asNum("max-mthreads");	
-  checkOrFixFDS();
 
   if(g_numThreads == 1) {
     L<<Logger::Warning<<"Operating unthreaded"<<endl;
@@ -1968,22 +2105,30 @@ try
       *t_pdl = shared_ptr<RecursorLua>(new RecursorLua(::arg()["lua-dns-script"]));
       L<<Logger::Warning<<"Loaded 'lua' script from '"<<::arg()["lua-dns-script"]<<"'"<<endl;
     }
-    
   }
   catch(std::exception &e) {
     L<<Logger::Error<<"Failed to load 'lua' script from '"<<::arg()["lua-dns-script"]<<"': "<<e.what()<<endl;
-    exit(99);
+    _exit(99);
   }
   
   t_traceRegex = new shared_ptr<Regex>();
-  
-  
-  t_remotes = new RemoteKeeper();
-  t_remotes->remotes.resize(::arg().asNum("remotes-ringbuffer-entries") / g_numThreads); 
-  
-  if(!t_remotes->remotes.empty())
-    memset(&t_remotes->remotes[0], 0, t_remotes->remotes.size() * sizeof(RemoteKeeper::remotes_t::value_type));
-  
+  unsigned int ringsize=::arg().asNum("stats-ringbuffer-entries") / g_numWorkerThreads;
+  if(ringsize) {
+    t_remotes = new addrringbuf_t();
+    if(g_weDistributeQueries)  // if so, only 1 thread does recvfrom
+      t_remotes->set_capacity(::arg().asNum("stats-ringbuffer-entries"));   
+    else
+      t_remotes->set_capacity(ringsize);   
+    t_servfailremotes = new addrringbuf_t();
+    t_servfailremotes->set_capacity(ringsize);   
+    t_largeanswerremotes = new addrringbuf_t();
+    t_largeanswerremotes->set_capacity(ringsize);   
+
+    t_queryring = new boost::circular_buffer<pair<string, uint16_t> >();
+    t_queryring->set_capacity(ringsize);   
+    t_servfailqueryring = new boost::circular_buffer<pair<string, uint16_t> >();
+    t_servfailqueryring->set_capacity(ringsize);   
+  }
   
   MT=new MTasker<PacketID,string>(::arg().asNum("stack-size"));
   
@@ -2020,7 +2165,7 @@ try
 
   time_t last_carbon=0;
   time_t carbonInterval=::arg().asNum("carbon-interval");
-  counter=0; // used to periodically execute certain tasks
+  counter=AtomicCounter(0); // used to periodically execute certain tasks
   for(;;) {
     while(MT->schedule(&g_now)); // MTasker letting the mthreads do their thing
       
@@ -2102,7 +2247,6 @@ int main(int argc, char **argv)
   try {
     ::arg().set("stack-size","stack size per mthread")="200000";
     ::arg().set("soa-minimum-ttl","Don't change")="0";
-    ::arg().set("soa-serial-offset","Don't change")="0";
     ::arg().set("no-shuffle","Don't change")="off";
     ::arg().set("local-port","port to listen on")="53";
     ::arg().set("local-address","IP addresses to listen on, separated by spaces or commas. Also accepts ports.")="127.0.0.1";
@@ -2122,6 +2266,7 @@ int main(int argc, char **argv)
     ::arg().set("experimental-webserver-address", "IP Address of webserver to listen on") = "127.0.0.1";
     ::arg().set("experimental-webserver-port", "Port of webserver to listen on") = "8082";
     ::arg().set("experimental-webserver-password", "Password required for accessing the webserver") = "";
+    ::arg().set("webserver-allow-from","Webserver access is only allowed from these subnets")="0.0.0.0/0,::/0";
     ::arg().set("experimental-api-config-dir", "Directory where REST API stores config and zones") = "";
     ::arg().set("experimental-api-key", "REST API Static authentication key (required for API use)") = "";
     ::arg().set("carbon-ourname", "If set, overrides our reported hostname for carbon stats")="";
@@ -2152,12 +2297,12 @@ int main(int argc, char **argv)
     ::arg().set("max-packetcache-entries", "maximum number of entries to keep in the packetcache")="500000";
     ::arg().set("packetcache-servfail-ttl", "maximum number of seconds to keep a cached servfail entry in packetcache")="60";
     ::arg().set("server-id", "Returned when queried for 'server.id' TXT or NSID, defaults to hostname")="";
-    ::arg().set("remotes-ringbuffer-entries", "maximum number of packets to store statistics for")="0";
+    ::arg().set("stats-ringbuffer-entries", "maximum number of packets to store statistics for")="10000";
     ::arg().set("version-string", "string reported on version.pdns or version.bind")=fullVersionString();
     ::arg().set("allow-from", "If set, only allow these comma separated netmasks to recurse")=LOCAL_NETS;
     ::arg().set("allow-from-file", "If set, load allowed netmasks from this file")="";
     ::arg().set("entropy-source", "If set, read entropy from this file")="/dev/urandom";
-    ::arg().set("dont-query", "If set, do not query these netmasks for DNS data")=LOCAL_NETS; 
+    ::arg().set("dont-query", "If set, do not query these netmasks for DNS data")=DONT_QUERY; 
     ::arg().set("max-tcp-per-client", "If set, maximum number of TCP sessions per client (IP address)")="0";
     ::arg().set("spoof-nearmiss-max", "If non-zero, assume spoofing after this many near misses")="20";
     ::arg().set("single-socket", "If set, only use a single socket for outgoing queries")="off";
@@ -2174,10 +2319,13 @@ int main(int argc, char **argv)
 //    ::arg().setSwitch( "disable-edns-ping", "Disable EDNSPing - EXPERIMENTAL, LEAVE DISABLED" )= "no"; 
     ::arg().setSwitch( "disable-edns", "Disable EDNS - EXPERIMENTAL, LEAVE DISABLED" )= ""; 
     ::arg().setSwitch( "disable-packetcache", "Disable packetcache" )= "no"; 
-    ::arg().setSwitch( "pdns-distributes-queries", "If PowerDNS itself should distribute queries over threads (EXPERIMENTAL)")="no";
+    ::arg().setSwitch( "pdns-distributes-queries", "If PowerDNS itself should distribute queries over threads")="";
+    ::arg().setSwitch( "root-nx-trust", "If set, believe that an NXDOMAIN from the root means the TLD does not exist")="no";
     ::arg().setSwitch( "any-to-tcp","Answer ANY queries with tc=1, shunting to TCP" )="no";
     ::arg().set("udp-truncation-threshold", "Maximum UDP response size before we truncate")="1680";
     ::arg().set("minimum-ttl-override", "Set under adverse conditions, a minimum TTL")="0";
+    ::arg().set("max-qperq", "Maximum outgoing queries per query")="50";
+    ::arg().set("max-total-msec", "Maximum total wall-clock time per query in milliseconds, 0 for unlimited")="7000";
 
     ::arg().set("include-dir","Include *.conf files from this directory")="";
     ::arg().set("security-poll-suffix","Domain name from which to query security update notifications")="secpoll.powerdns.com.";
@@ -2207,6 +2355,9 @@ int main(int argc, char **argv)
 
     ::arg().set("delegation-only")=toLower(::arg()["delegation-only"]);
 
+    if(::arg().asNum("threads")==1)
+      ::arg().set("pdns-distributes-queries")="no";
+
     if(::arg().mustDo("help")) {
       cout<<"syntax:"<<endl<<endl;
       cout<<::arg().helpstring(::arg()["help"])<<endl;
@@ -2219,8 +2370,12 @@ int main(int argc, char **argv)
     }
 
     Logger::Urgency logUrgency = (Logger::Urgency)::arg().asNum("loglevel");
+
     if (logUrgency < Logger::Error)
       logUrgency = Logger::Error;
+    if(!g_quiet && logUrgency < Logger::Info) { // Logger::Info=6, Logger::Debug=7
+      logUrgency = Logger::Info;                // if you do --quiet=no, you need Info to also see the query log
+    }
     L.setLoglevel(logUrgency);
     L.toConsole(logUrgency);
 
