@@ -42,6 +42,7 @@ try
     if(p) *p=0;
 
     hostname=tmp;
+    boost::replace_all(hostname, ".", "_");    
   }
   BOOST_FOREACH(const all_t::value_type& val, all) {
     str<<"pdns."<<hostname<<".recursor."<<val.first<<' '<<val.second<<' '<<now<<"\r\n";
@@ -52,7 +53,7 @@ try
   if(ret < 0)
     L<<Logger::Warning<<"Error writing carbon data to "<<remote.toStringWithPort()<<": "<<strerror(errno)<<endl;
   if(ret==0)
-    L<<Logger::Warning<<"Timeout connecting/writing carbon data to "<<remote.toStringWithPort();
+    L<<Logger::Warning<<"Timeout connecting/writing carbon data to "<<remote.toStringWithPort()<<endl;
  }
 catch(PDNSException& e)
 {
