@@ -12,6 +12,7 @@
 #include <boost/format.hpp>
 #include <boost/algorithm/string.hpp>
 #include <boost/foreach.hpp>
+#include "version.hh"
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -23,6 +24,8 @@
 #include "lock.hh"
 #include "responsestats.hh"
 
+#include "secpoll-recursor.hh"
+#include "pubsuffix.hh"
 #include "namespaces.hh"
 pthread_mutex_t g_carbon_config_lock=PTHREAD_MUTEX_INITIALIZER;
 
@@ -508,6 +511,7 @@ RecursorControlParser::RecursorControlParser()
 
   addGetStat("client-parse-errors", &g_stats.clientParseError);
   addGetStat("server-parse-errors", &g_stats.serverParseError);
+  addGetStat("too-old-drops", &g_stats.tooOldDrops);
 
   addGetStat("answers0-1", &g_stats.answers0_1);
   addGetStat("answers1-10", &g_stats.answers1_10);
@@ -536,6 +540,7 @@ RecursorControlParser::RecursorControlParser()
   addGetStat("failed-host-entries", boost::bind(getFailedHostsSize));
 
   addGetStat("concurrent-queries", boost::bind(getConcurrentQueries)); 
+  addGetStat("security-status", &g_security_status);
   addGetStat("outgoing-timeouts", &SyncRes::s_outgoingtimeouts);
   addGetStat("tcp-outqueries", &SyncRes::s_tcpoutqueries);
   addGetStat("all-outqueries", &SyncRes::s_outqueries);
@@ -582,30 +587,90 @@ static void doExit()
 
 static void doExitNicely()
 {
-  //extern void printCallers();
-  // printCallers();
   doExitGeneric(true);
 }
 
-vector<ComboAddress>* pleaseGetRemotes()
+vector<pair<string, uint16_t> >* pleaseGetQueryRing()
 {
-  return new vector<ComboAddress>(t_remotes->remotes);
+  typedef pair<string,uint16_t> query_t;
+  vector<query_t >* ret = new vector<query_t>();
+  if(!t_queryring)
+    return ret;
+  ret->reserve(t_queryring->size());
+
+  BOOST_FOREACH(const query_t& q, *t_queryring) {
+    ret->push_back(q);
+  }
+  return ret;
+}
+vector<pair<string,uint16_t> >* pleaseGetServfailQueryRing()
+{
+  typedef pair<string,uint16_t> query_t;
+  vector<query_t>* ret = new vector<query_t>();
+  if(!t_servfailqueryring)
+    return ret;
+  ret->reserve(t_queryring->size());
+  BOOST_FOREACH(const query_t& q, *t_servfailqueryring) {
+    ret->push_back(q);
+  }
+  return ret;
 }
 
-string doTopRemotes()
+
+
+typedef boost::function<vector<ComboAddress>*()> pleaseremotefunc_t;
+typedef boost::function<vector<pair<string,uint16_t> >*()> pleasequeryfunc_t;
+
+vector<ComboAddress>* pleaseGetRemotes()
+{
+  vector<ComboAddress>* ret = new vector<ComboAddress>();
+  if(!t_remotes)
+    return ret;
+
+  ret->reserve(t_remotes->size());
+  BOOST_FOREACH(const ComboAddress& ca, *t_remotes) {
+    ret->push_back(ca);
+  }
+  return ret;
+}
+
+vector<ComboAddress>* pleaseGetServfailRemotes()
+{
+  vector<ComboAddress>* ret = new vector<ComboAddress>();
+  if(!t_servfailremotes)
+    return ret;
+  ret->reserve(t_servfailremotes->size());
+  BOOST_FOREACH(const ComboAddress& ca, *t_servfailremotes) {
+    ret->push_back(ca);
+  }
+  return ret;
+}
+
+vector<ComboAddress>* pleaseGetLargeAnswerRemotes()
+{
+  vector<ComboAddress>* ret = new vector<ComboAddress>();
+  if(!t_largeanswerremotes)
+    return ret;
+  ret->reserve(t_largeanswerremotes->size());
+  BOOST_FOREACH(const ComboAddress& ca, *t_largeanswerremotes) {
+    ret->push_back(ca);
+  }
+  return ret;
+}
+
+string doGenericTopRemotes(pleaseremotefunc_t func)
 {
   typedef map<ComboAddress, int, ComboAddress::addressOnlyLessThan> counts_t;
   counts_t counts;
 
-  vector<ComboAddress> remotes=broadcastAccFunction<vector<ComboAddress> >(pleaseGetRemotes);
+  vector<ComboAddress> remotes=broadcastAccFunction<vector<ComboAddress> >(func);
     
   unsigned int total=0;
-  for(RemoteKeeper::remotes_t::const_iterator i = remotes.begin(); i != remotes.end(); ++i)
-    if(i->sin4.sin_family) {
-      total++;
-      counts[*i]++;
-    }
-
+  BOOST_FOREACH(const ComboAddress& ca, remotes) {
+    total++;
+    counts[ca]++;
+  }
+  
   typedef std::multimap<int, ComboAddress> rcounts_t;
   rcounts_t rcounts;
   
@@ -613,13 +678,104 @@ string doTopRemotes()
     rcounts.insert(make_pair(-i->second, i->first));
 
   ostringstream ret;
-  ret<<"Over last "<<total<<" queries:\n";
+  ret<<"Over last "<<total<<" entries:\n";
   format fmt("%.02f%%\t%s\n");
-  int limit=0;
-  if(total)
-    for(rcounts_t::const_iterator i=rcounts.begin(); i != rcounts.end() && limit < 20; ++i, ++limit)
+  int limit=0, accounted=0;
+  if(total) {
+    for(rcounts_t::const_iterator i=rcounts.begin(); i != rcounts.end() && limit < 20; ++i, ++limit) {
       ret<< fmt % (-100.0*i->first/total) % i->second.toString();
+      accounted+= -i->first;
+    }
+    ret<< '\n' << fmt % (100.0*(total-accounted)/total) % "rest";
+  }
+  return ret.str();
+}
 
+namespace {
+  typedef vector<vector<string> > pubs_t;
+  pubs_t g_pubs;
+}
+
+void sortPublicSuffixList()
+{
+  for(const char** p=&g_pubsuffix; *p; ++p) {
+    string low=toLower(*p);
+
+    vector<string> parts;
+    stringtok(parts, low, ".");
+    reverse(parts.begin(), parts.end());
+    g_pubs.push_back(parts);
+  }
+  sort(g_pubs.begin(), g_pubs.end());
+}
+
+string getRegisteredName(const std::string& dom)
+{
+  vector<string> parts;
+  stringtok(parts, dom, ".");
+  if(parts.size()<=2)
+    return dom;
+  reverse(parts.begin(), parts.end());
+  BOOST_FOREACH(string& str, parts) { str=toLower(str); };
+
+  // uk co migweb 
+  string last;
+  while(!parts.empty()) {
+    if(parts.size()==1 || binary_search(g_pubs.begin(), g_pubs.end(), parts)) {
+  
+      string ret=last;
+      if(!ret.empty())
+	ret+=".";
+      
+      BOOST_REVERSE_FOREACH(const std::string& p, parts) {
+	ret+=p+".";
+      }
+      return ret;
+    }
+
+    last=parts[parts.size()-1];
+    parts.resize(parts.size()-1);
+  }
+  return "??";
+}
+
+static string nopFilter(const std::string& str)
+{
+  return str;
+}
+
+string doGenericTopQueries(pleasequeryfunc_t func, boost::function<string(const std::string&)> filter=nopFilter)
+{
+  typedef pair<string,uint16_t> query_t;
+  typedef map<query_t, int> counts_t;
+  counts_t counts;
+  vector<query_t> queries=broadcastAccFunction<vector<query_t> >(func);
+    
+  unsigned int total=0;
+  BOOST_FOREACH(const query_t& q, queries) {
+    total++;
+    counts[make_pair(toLower(filter(q.first)),q.second)]++;
+  }
+
+  typedef std::multimap<int, query_t> rcounts_t;
+  rcounts_t rcounts;
+  
+  for(counts_t::const_iterator i=counts.begin(); i != counts.end(); ++i)
+    rcounts.insert(make_pair(-i->second, i->first));
+
+  ostringstream ret;
+  ret<<"Over last "<<total<<" entries:\n";
+  format fmt("%.02f%%\t%s\n");
+  int limit=0, accounted=0;
+  if(total) {
+    for(rcounts_t::const_iterator i=rcounts.begin(); i != rcounts.end() && limit < 20; ++i, ++limit) {
+      ret<< fmt % (-100.0*i->first/total) % (i->second.first+"|"+DNSRecordContent::NumberToType(i->second.second));
+      accounted+= -i->first;
+    }
+    ret<< '\n' << fmt % (100.0*(total-accounted)/total) % "rest";
+  }
+
+  
   return ret.str();
 }
 
@@ -662,8 +818,13 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
 "set-minimum-ttl value            set mininum-ttl-override\n"
 "set-carbon-server                set a carbon server for telemetry\n"
 "trace-regex [regex]              emit resolution trace for matching queries (empty regex to clear trace)\n"
+"top-largeanswer-remotes          show top remotes receiving large answers\n"
+"top-queries                      show top queries\n"
 "top-remotes                      show top remotes\n"
+"top-servfail-queries             show top queries receiving servfail answers\n"
+"top-servfail-remotes             show top remotes receiving servfail answers\n"
 "unload-lua-script                unload Lua script\n"
+"version                          return Recursor version number\n"
 "wipe-cache domain0 [domain1] ..  wipe domain data from cache\n";
 
   if(cmd=="get-all")
@@ -679,6 +840,10 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
     *command=&doExit;
     return "bye\n";
   }
+
+  if(cmd=="version") {
+    return getPDNSVersion()+"\n";
+  }
   
   if(cmd=="quit-nicely") {
     *command=&doExitNicely;
@@ -731,7 +896,27 @@ string RecursorControlParser::getAnswer(const string& question, RecursorControlP
 
 
   if(cmd=="top-remotes")
-    return doTopRemotes();
+    return doGenericTopRemotes(pleaseGetRemotes);
+
+  if(cmd=="top-queries")
+    return doGenericTopQueries(pleaseGetQueryRing);
+
+  if(cmd=="top-pub-queries")
+    return doGenericTopQueries(pleaseGetQueryRing, getRegisteredName);
+
+  if(cmd=="top-servfail-queries")
+    return doGenericTopQueries(pleaseGetServfailQueryRing);
+
+  if(cmd=="top-pub-servfail-queries")
+    return doGenericTopQueries(pleaseGetServfailQueryRing, getRegisteredName);
+
+
+  if(cmd=="top-servfail-remotes")
+    return doGenericTopRemotes(pleaseGetServfailRemotes);
+
+  if(cmd=="top-largeanswer-remotes")
+    return doGenericTopRemotes(pleaseGetLargeAnswerRemotes);
+
 
   if(cmd=="current-queries")
     return doCurrentQueries();
