@@ -23,15 +23,16 @@
 #include "logger.hh"
 #include "version.hh"
 #include "version_generated.h"
+#include <polarssl/version.h>
 
 static ProductType productType;
 
 string compilerVersion()
 {
 #if defined(__clang__)
-  return string("clang "__clang_version__);
+  return string("clang " __clang_version__ );
 #elif defined(__GNUC__)
-  return string("gcc "__VERSION__);
+  return string("gcc " __VERSION__ );
 #else  // add other compilers here
   return string("Unknown compiler");
 #endif
@@ -48,6 +49,11 @@ string productName() {
   return "Unknown";
 }
 
+string getPDNSVersion()
+{
+  return PDNS_VERSION;
+}
+
 // REST API product type
 string productTypeApiType() {
   switch (productType) {
@@ -61,10 +67,10 @@ string productTypeApiType() {
 
 void showProductVersion()
 {
-  theL()<<Logger::Warning<<productName()<<" "<<PDNS_VERSION<<" ("DIST_HOST") "
-    "(C) 2001-2014 PowerDNS.COM BV" << endl;
+  theL()<<Logger::Warning<<productName()<<" "<< PDNS_VERSION <<" (" DIST_HOST ") "
+    "(C) 2001-2015 PowerDNS.COM BV" << endl;
   theL()<<Logger::Warning<<"Using "<<(sizeof(unsigned long)*8)<<"-bits mode. "
-    "Built on "BUILD_DATE" by "BUILD_HOST", "<<compilerVersion()<<"."<<endl;
+    "Built on " BUILD_DATE " by " BUILD_HOST ", "<<compilerVersion()<<"."<<endl;
   theL()<<Logger::Warning<<"PowerDNS comes with ABSOLUTELY NO WARRANTY. "
     "This is free software, and you are welcome to redistribute it "
     "according to the terms of the GPL version 2." << endl;
@@ -82,14 +88,17 @@ void showBuildConfiguration()
 #ifdef HAVE_CRYPTOPP
     "cryptopp " <<
 #endif
+#ifdef HAVE_ED25519
+    "ed25519 " <<
+#endif
 #ifdef HAVE_LIBDL
     "libdl " <<
 #endif
 #ifdef HAVE_LUA
     "lua " <<
 #endif
-#ifdef REMOTEBACKEND_HTTP
-    "remotebackend-http" <<
+#ifdef REMOTEBACKEND_ZEROMQ
+    "remotebackend-zeromq" <<
 #endif
 #ifdef VERBOSELOG
     "verboselog" <<
@@ -99,12 +108,15 @@ void showBuildConfiguration()
   // Auth only
   theL()<<Logger::Warning<<"Built-in modules: "<<PDNS_MODULES<<endl;
 #endif
+#ifndef POLARSSL_SYSTEM
+  theL()<<Logger::Warning<<"Built-in PolarSSL: "<<POLARSSL_VERSION_STRING<<endl;
+#endif
 }
 
 string fullVersionString()
 {
   ostringstream s;
-  s<<productName()<<" "PDNS_VERSION" ("DIST_HOST" built "BUILD_DATE" "BUILD_HOST")";
+  s<<productName()<<" " PDNS_VERSION " (" DIST_HOST " built " BUILD_DATE " " BUILD_HOST ")";
   return s.str();
 }
 
